package materials

import Hit
import math.Ray
import math.Vector
import things.Scene
import kotlin.math.cos
import kotlin.math.ln
import kotlin.math.log
import kotlin.math.sqrt
import kotlin.random.Random

data class PathTracedMaterial(
    val color: MaterialColor,
    val emissionColor: MaterialColor = MaterialColor(0f, 0f, 0f),
    val specularProbability: Float = 0f,
    val smoothness: Float = 0.5f,
    val emissionStrength: Float = 0f,
    val specularColour: MaterialColor,
) : Material {
    override fun shade(ray: Ray, hit: Hit, scene: Scene): MaterialColor {
        // source: https://github.com/SebLague/Ray-Tracing/blob/main/Assets/Scripts/Shaders/RayTracer.shader#L328

        var incomingLight = MaterialColor.Black
        var rayColor = MaterialColor.White

        var hit: Hit? = hit
        var ray = ray

        repeat(maxBounces) {
            if (hit != null && hit.material is PathTracedMaterial) {
                val isSpecular = hit.material.specularProbability >= Random.Default.nextFloat()

                val diffuseDir = (hit.normal + Random.Default.direction()).normalized()
                val specularDir = reflect(ray.direction, hit.normal)

                val rayDir = if (isSpecular) {
                    lerp(diffuseDir, specularDir, hit.material.smoothness).normalized()
                } else {
                    diffuseDir
                }

                val emittedLight = hit.material.emissionColor * hit.material.emissionStrength;
                incomingLight += emittedLight * rayColor;
                rayColor *= if (isSpecular) hit.material.specularColour else hit.material.color

                val p = maxOf(rayColor.r, rayColor.g, rayColor.b)
                if (Random.Default.nextFloat() >= p) {
                    return@repeat
                }

                rayColor *= 1.0f / p;

                ray = Ray(hit.point, rayDir)

            } else if (hit != null) {
                incomingLight += hit.material.shade(ray, hit, scene) * rayColor
                return@repeat
            } else {
                incomingLight += scene.ambientLight * rayColor
                return@repeat
            }

            hit = scene.intersects(ray)
        }

        return incomingLight
    }

}

const val maxBounces = 4

fun Random.normalDistributedFloat(): Float {
    val theta = nextFloat() * 2 * Math.PI.toFloat()
    val rho = sqrt(-2 * ln(nextFloat()))
    return rho * cos(theta)
}

fun Random.direction(): Vector =
    Vector(normalDistributedFloat(), normalDistributedFloat(), normalDistributedFloat()).normalized()

fun lerp(a: Float, b: Float, c: Float): Float = a * (1 - c) + b * c
fun lerp(a: MaterialColor, b: MaterialColor, c: Float): MaterialColor = a * (1 - c) + b * c
fun lerp(a: Vector, b: Vector, c: Float): Vector = a * (1 - c) + b * c
